/*
 * Copyright (C) 2013-2018 CERN (www.cern.ch)
 * Author: Federico Vaga <federico.vaga@cern.ch>
 * Author: Tomasz Wlostowski <tomasz.wlostowski@cern.ch>
 *
 * SPDX-License-Identifier: LGPL-3.0-or-later
 */

#include <string.h>

#include "mockturtle-rt.h"
#include <mockturtle-framework.h>
#include "wrtd-common.h"
#include "hw/fmctdc-direct.h"
#include "hw/tdc_regs.h"

#ifdef SIMULATION
static const int is_simulation = 1;
#else
static const int is_simulation = 0;
#endif

#define NBR_CPUS 2
#define CPU_IDX 0

#define NBR_LOCAL_INPUTS 5
#define NBR_LOCAL_OUTPUTS 0
#define NBR_RULES 16
#define NBR_DEVICES 1
#define NBR_ALARMS 0
#define DEVICES_IO_ADDR { TRTL_ADDR_DP_BASE, 0, 0, 0}
#define APP_ID WRTD_APP_TDC_5CH
#define APP_VER RT_VERSION(2, 0)
#define APP_NAME "wrtd-tdc"
#define WRTD_NET_TX

#include "wrtd-rt-common.h"

#include "wrtd-tdc.c"

static struct wrtd_tdc_dev tdc0 =
{
        .io_addr = 0x0,
};

static inline int wr_link_up(void)
{
        return tdc_wr_link_up(&tdc0);
}

static inline int wr_time_locked(void)
{
        return tdc_wr_time_locked(&tdc0);
}

static inline int wr_time_ready(void)
{
        return tdc_wr_time_ready(&tdc0);
}

static inline void wr_enable_lock(int enable)
{
        tdc_wr_enable_lock(&tdc0, enable);
}

static uint32_t tai_start = 0;


static void wrtd_local_output(struct wrtd_event *ev, unsigned ch)
{
        return;
}


/**
 * It updates the White-Rabbit link status
 */
static void wr_update_link(void)
{
	switch (root.wr_state) {
	case WR_LINK_OFFLINE:
		if (wr_link_up())
			root.wr_state = WR_LINK_ONLINE;
		break;
	case WR_LINK_ONLINE:
		if (wr_time_ready()) {
			root.wr_state = WR_LINK_SYNCING;
			wr_enable_lock(1);
		}
		break;
	case WR_LINK_SYNCING:
		if (wr_time_locked()) {
			pr_debug("rt-tdc: WR synced, waiting for TDC plumbing to catch up...\n\r");
			root.wr_state = WR_LINK_WAIT;
			tai_start = lr_readl(MT_CPU_LR_REG_TAI_SEC);
		}
		break;
	case WR_LINK_WAIT:
		if (is_simulation
                    || (lr_readl(MT_CPU_LR_REG_TAI_SEC) >= (tai_start + 4))) {
			pr_debug("rt-tdc: WR TDC synced\n\r");
			root.wr_state = WR_LINK_SYNCED;
		}
		break;
	case WR_LINK_SYNCED:
		break;
	}

	if (root.wr_state != WR_LINK_OFFLINE && !wr_link_up()) {
		pr_error("rt-tdc: WR sync lost\n\r");
		root.wr_state = WR_LINK_OFFLINE;
		wr_enable_lock(0);
	}
}

static int wrtd_i_sim_init(struct wrtd_tdc_dev *tdc)
{
#if 0
	int i;

	for(i = 0; i < TDC_NUM_CHANNELS; i++) {
		wrtd_in_channels[i].config.log_level = WRTD_LOG_ALL;
		wrtd_in_channels[i].config.id.system = 0;
		wrtd_in_channels[i].config.id.source_port = i;
		wrtd_in_channels[i].config.id.trigger = i;
		wrtd_in_channels[i].config.flags =
		  WRTD_ENABLED | WRTD_ARMED | WRTD_TRIGGER_ASSIGNED;
	}
#endif

	tdc_writel(tdc, 0x40, BASE_DP_TDC_DIRECT + DR_REG_DEAD_TIME);
	tdc_writel(tdc, 0x1f, BASE_DP_TDC_DIRECT + DR_REG_CHAN_ENABLE);

	return 0;
}

static int wrtd_user_init(void)
{
	if (!is_simulation)
		tdc_init(&tdc0);

	tdc_wr_enable_lock(&tdc0, 0);

	/* Skip WR sync and automatically generate some events when
           simulating */
	if (is_simulation)
		wrtd_i_sim_init(&tdc0);

	pr_debug("rt-input firmware initialized.\n\r");

	return 0;
}

static void wrtd_io(void)
{
        tdc_input(&tdc0);
}
